import { world, ItemStack, MolangVariableMap, GameMode } from "@minecraft/server";
import { MinecraftFlowerBlockTag, MinecraftSoulBlockTag, MinecraftOreBlockTag, MinecraftFenceBlockTag, MinecraftStairBlockTag } from "./blocktags";

export const StructurePattern = [
    /**
     * $ $
     *  #
     * $ $
     */
    { offset: { x: 0, y: -1, z: 0 }, blockType: MinecraftFenceBlockTag },

    { offset: { x: 1, y: -1, z: 1 }, blockType: MinecraftFlowerBlockTag },
    { offset: { x: 1, y: -1, z: -1 }, blockType: MinecraftFlowerBlockTag },
    { offset: { x: -1, y: -1, z: 1 }, blockType: MinecraftFlowerBlockTag },
    { offset: { x: -1, y: -1, z: -1 }, blockType: MinecraftFlowerBlockTag },
    
    /**
     * ###
     * ###
     * ###
     */
    { offset: { x: 0, y: -2, z: 0 }, blockType: MinecraftOreBlockTag },
    { offset: { x: 1, y: -2, z: 0 }, blockType: MinecraftStairBlockTag },
    { offset: { x: -1, y: -2, z: 0 }, blockType: MinecraftStairBlockTag },
    { offset: { x: 0, y: -2, z: 1 }, blockType: MinecraftStairBlockTag },
    { offset: { x: 0, y: -2, z: -1 }, blockType: MinecraftStairBlockTag },

    { offset: { x: 1, y: -2, z: 1 }, blockType: MinecraftSoulBlockTag },
    { offset: { x: 1, y: -2, z: -1 }, blockType: MinecraftSoulBlockTag },
    { offset: { x: -1, y: -2, z: 1 }, blockType: MinecraftSoulBlockTag },
    { offset: { x: -1, y: -2, z: -1 }, blockType: MinecraftSoulBlockTag }
];
export const FailedStructurePattern = [
    /**
     *  
     *  #
     *  
     */
    { offset: { x: 0, y: -1, z: 0 }, blockType: MinecraftFenceBlockTag },
    
    /**
     * ###
     * ###
     * ###
     */
    { offset: { x: 0, y: -2, z: 0 }, blockType: MinecraftOreBlockTag },
    { offset: { x: 1, y: -2, z: 0 }, blockType: MinecraftStairBlockTag },
    { offset: { x: -1, y: -2, z: 0 }, blockType: MinecraftStairBlockTag },
    { offset: { x: 0, y: -2, z: 1 }, blockType: MinecraftStairBlockTag },
    { offset: { x: 0, y: -2, z: -1 }, blockType: MinecraftStairBlockTag },

    { offset: { x: 1, y: -2, z: 1 }, blockType: MinecraftSoulBlockTag },
    { offset: { x: 1, y: -2, z: -1 }, blockType: MinecraftSoulBlockTag },
    { offset: { x: -1, y: -2, z: 1 }, blockType: MinecraftSoulBlockTag },
    { offset: { x: -1, y: -2, z: -1 }, blockType: MinecraftSoulBlockTag }
];

function getBlockPatternResult(origin, pattern) {
    const shape = [];
    for (const { offset, blockType } of pattern) {
        const blockLocation = { x: origin.location.x + offset.x, y: origin.location.y + offset.y, z: origin.location.z + offset.z };
        const block = origin.dimension.getBlock(blockLocation);
        if (!block)
            return false;
        let matches = false;
        for (const string of blockType) {
            if (block.typeId === string) {
                matches = true;
                break;
            }
        }
        if (!matches)
            return false;
        
        shape.push(block);
    }
    return shape;
}

function centreLocation(location) {
    return {
        x: location.x + 0.5,
        y: location.y + 0.5,
        z: location.z + 0.5
    };
}
function sendError(dimension, location, player, message) { // Fix errors
    player.sendMessage(message);
    dimension.playSound('random.fizz', centreLocation(location), { pitch: 2.1, volume: 0.5 });
    dimension.spawnParticle('minecraft:basic_smoke_particle', centreLocation(location), new MolangVariableMap());
}

world.afterEvents.playerPlaceBlock.subscribe(({ player, block, dimension }) => {
    if (block.typeId !== 'minecraft:skull')  // ID: 3
        return;

    const headBlock = block;
    const headBlockPos = headBlock.location;
    const result = getBlockPatternResult(headBlock, StructurePattern);

    if (result) {
        let deadPlayer;
        dimension.getPlayers({
            location: headBlockPos,
            farthest: dimension.getPlayers().length
        }).forEach((player) => {
            if (player.getGameMode() === GameMode.spectator)
                deadPlayer = player;
        });

        if (deadPlayer) {
            const groundY = result.reduce((a, b) => a < b.y ? a : b.y, headBlockPos.y);
            const groundPos = { x: centreLocation(headBlockPos).x , y: groundY +1, z: centreLocation(headBlockPos).z };
            for (const block of result) {
                if (groundPos.y <= block.location.y)
                    block?.setType('minecraft:air');
            }
            headBlock.setType('minecraft:air');
            dimension.spawnEntity('minecraft:lightning_bolt', groundPos);
            deadPlayer?.teleport(groundPos, { dimension: dimension, facingLocation: player.location, keepVelocity: false });
            deadPlayer?.addTag('hardcore_revive');

        } else {
            sendError(dimension, headBlockPos, player, {translate:'hardcorerevive.error_message.absent'});
        }
    } else {
        const result2 = getBlockPatternResult(headBlock, FailedStructurePattern);
        if (result2)
            sendError(dimension, headBlockPos, player, {translate:'hardcorerevive.error_message.wrong'});
    }
});

world.afterEvents.entityDie.subscribe(({damageSource, deadEntity}) => {
    if (deadEntity.typeId !== 'minecraft:player')
        return;

    const player = deadEntity;
    const deathLocation = player.location
    const dimension = player.dimension;
    const playerHead = new ItemStack('minecraft:skull', 1);

    playerHead.nameTag = `§r§e${player.nameTag}'s Skull Head`; //Translate
    playerHead.setLore(['§r§4Used to revive this player§r']);

    dimension.spawnItem(playerHead, centreLocation(deathLocation));
    player.setGameMode(GameMode.spectator);
    
    /*
    const headItem = dimension.getEntities({type: 'minecraft:item', name: playerHead.nameTag, location: deathLocation, maxDistance: 20})[0];
    headItem.setDynamicProperty('SkullOwner', player.nameTag);
    console.warn(headItem.getDynamicProperty('SkullOwner'));
    */

    //Death Coords
    player.sendMessage({translate: 'hardcorerevive.death_location.message', with: [`${Math.floor(deathLocation.x)}`, `${Math.floor(deathLocation.y)}`, `${Math.floor(deathLocation.z)}`]});
    player.sendMessage({translate: 'hardcorerevive.death_location.dimension', with: [dimension.id]});
});